/*
Mouse XY inverter driver
driver replace X and Y coordinates of pointer

writed by rpy3uH 
E-mail: gruzino.abhaz@gmail.com
Last modified: 29.05.2001
*/

#include <ntddk.h>
#include <ntddmou.h>

typedef struct _DEVICE_EXTENSION {
    PDEVICE_OBJECT  TopOfStack;
} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

NTSTATUS
mousinvAddDevice(
    IN PDRIVER_OBJECT Driver,
    IN PDEVICE_OBJECT PDO)
{
    PDEVICE_EXTENSION   devExt;
    IO_ERROR_LOG_PACKET errorLogEntry;
    PDEVICE_OBJECT      device;
    NTSTATUS            status = STATUS_SUCCESS;

    // Create a filter device and attach it to the device stack.
    status = IoCreateDevice(Driver, sizeof(DEVICE_EXTENSION), 
                            NULL, FILE_DEVICE_KEYBOARD,   
                            0, FALSE, &device);

    if (!NT_SUCCESS(status)) return status;

    RtlZeroMemory(device->DeviceExtension, sizeof(DEVICE_EXTENSION));

    devExt = (PDEVICE_EXTENSION) device->DeviceExtension;
    devExt->TopOfStack = IoAttachDeviceToDeviceStack(device, PDO);

    device->Flags |= (DO_BUFFERED_IO | DO_POWER_PAGABLE);
    device->Flags &= ~DO_DEVICE_INITIALIZING;
    return status;
}

NTSTATUS mousinvPnP(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp)
{
    PDEVICE_EXTENSION           devExt; 
    PIO_STACK_LOCATION          irpStack;
    NTSTATUS                    status = STATUS_SUCCESS;
    KIRQL                       oldIrql;
    KEVENT                      event;        

    devExt = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    irpStack = IoGetCurrentIrpStackLocation(Irp);

    switch (irpStack->MinorFunction) {
    case IRP_MN_REMOVE_DEVICE:
        
        //
        // Detach from the target device after passing the IRP
        // down the devnode stack.
        //
        IoSkipCurrentIrpStackLocation(Irp);
        IoCallDriver(devExt->TopOfStack, Irp);

        IoDetachDevice(devExt->TopOfStack); 
        IoDeleteDevice(DeviceObject);

        status = STATUS_SUCCESS;
        break;

    case IRP_MN_SURPRISE_REMOVAL:

        // Same as a remove device, but don't call IoDetach or IoDeleteDevice.
        IoSkipCurrentIrpStackLocation(Irp);
        status = IoCallDriver(devExt->TopOfStack, Irp);
        break;

    case IRP_MN_START_DEVICE: 
    case IRP_MN_QUERY_REMOVE_DEVICE:
    case IRP_MN_QUERY_STOP_DEVICE:
    case IRP_MN_CANCEL_REMOVE_DEVICE:
    case IRP_MN_CANCEL_STOP_DEVICE:
    case IRP_MN_FILTER_RESOURCE_REQUIREMENTS: 
    case IRP_MN_STOP_DEVICE:
    case IRP_MN_QUERY_DEVICE_RELATIONS:
    case IRP_MN_QUERY_INTERFACE:
    case IRP_MN_QUERY_CAPABILITIES:
    case IRP_MN_QUERY_DEVICE_TEXT:
    case IRP_MN_QUERY_RESOURCES:
    case IRP_MN_QUERY_RESOURCE_REQUIREMENTS:
    case IRP_MN_READ_CONFIG:
    case IRP_MN_WRITE_CONFIG:
    case IRP_MN_EJECT:
    case IRP_MN_SET_LOCK:
    case IRP_MN_QUERY_ID:
    case IRP_MN_QUERY_PNP_DEVICE_STATE:
    default:
        // Pass these through untouched
        IoSkipCurrentIrpStackLocation(Irp);
        status = IoCallDriver(devExt->TopOfStack, Irp);
        break;
    }

    return status;
}

NTSTATUS mousinvPower(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp)
{
    PDEVICE_EXTENSION   devExt;
    
    devExt = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    PoStartNextPowerIrp(Irp); 
    IoSkipCurrentIrpStackLocation(Irp);
    
    return PoCallDriver( devExt->TopOfStack, Irp );
}

NTSTATUS mousinvReadComplete( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp,
    IN PVOID Context)
{
    PIO_STACK_LOCATION  IrpSp;
    PMOUSE_INPUT_DATA mouseData;
    LONG numRecs, i;

    IrpSp = IoGetCurrentIrpStackLocation(Irp);
    if(NT_SUCCESS(Irp->IoStatus.Status)) 
    {
        mouseData = Irp->AssociatedIrp.SystemBuffer;
        numRecs = Irp->IoStatus.Information / sizeof(MOUSE_INPUT_DATA);

        for( i = 0; i < numRecs; i++ ) 
        {
        	LONG tmp = mouseData[i].LastX;        	
        	mouseData[i].LastX = mouseData[i].LastY;
        	mouseData[i].LastY = tmp;
        }
    }
    
    if (Irp->PendingReturned) IoMarkIrpPending( Irp );

    return Irp->IoStatus.Status;
}

NTSTATUS 
mousinvDispatchRead( 
    IN PDEVICE_OBJECT DeviceObject, 
    IN PIRP Irp )
{
    PDEVICE_EXTENSION   devExt;
    PIO_STACK_LOCATION  currentIrpStack;
    PIO_STACK_LOCATION  nextIrpStack;

    devExt = (PDEVICE_EXTENSION) DeviceObject->DeviceExtension;
    currentIrpStack = IoGetCurrentIrpStackLocation(Irp);
    nextIrpStack = IoGetNextIrpStackLocation(Irp);    

    *nextIrpStack = *currentIrpStack;
   
    IoSetCompletionRoutine( Irp, mousinvReadComplete, 
                            DeviceObject, TRUE, TRUE, TRUE );

    return IoCallDriver( devExt->TopOfStack, Irp );
}

NTSTATUS 
DriverDispatchGeneral(
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP           Irp)
{
    IoSkipCurrentIrpStackLocation(Irp);
    return IoCallDriver(((PDEVICE_EXTENSION) DeviceObject->DeviceExtension)->TopOfStack, Irp);
}

void DriverUnload(IN PDRIVER_OBJECT Driver)
{
    UNREFERENCED_PARAMETER(Driver);
}

NTSTATUS DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath)
{
    ULONG i;
     
    for (i = 0; i < IRP_MJ_MAXIMUM_FUNCTION; i++) 
    {
        DriverObject->MajorFunction[i] = DriverDispatchGeneral;
    }

    DriverObject->MajorFunction[IRP_MJ_READ] = mousinvDispatchRead;

    DriverObject->MajorFunction [IRP_MJ_POWER]  = mousinvPower;
    DriverObject->MajorFunction [IRP_MJ_PNP]  = mousinvPnP;
    DriverObject->DriverUnload = DriverUnload;
    DriverObject->DriverExtension->AddDevice = mousinvAddDevice;

    return STATUS_SUCCESS;
}